import 'package:hive/hive.dart' as h show Box;
import 'package:hive/hive.dart' hide Box;

class BoxCollection {
  final String name;
  final Set<String> boxNames;
  HiveCipher? _cipher;

  BoxCollection(this.name, this.boxNames);

  static Future<BoxCollection> open(
    String name,
    Set<String> boxNames, {
    String? path,
    List<int>? key,
  }) async {
    Hive.init(path ?? './');
    final collection = BoxCollection(name, boxNames);
    if (key != null) {
      collection._cipher = HiveAesCipher(key);
    }
    return collection;
  }

  Future<Box<V>> openBox<V>(String name, {bool preload = false}) async {
    if (!boxNames.contains(name)) {
      throw ('Box with name $name is not in the known box names of this collection.');
    }
    final box = Box<V>('${this.name}_$name', this);
    if (preload) {
      box._cachedBox = await Hive.openBox(box.name, encryptionCipher: _cipher);
    }
    return box;
  }

  List<Future<void> Function()>? _txnCache;

  Future<void> transaction(
    Future<void> Function() action, {
    List<String>? boxNames,
    bool readOnly = false,
  }) async {
    _txnCache = [];
    await action();
    final cache = List<Future<void> Function()>.from(_txnCache!);
    _txnCache = null;
    if (cache.isEmpty) return;
    await Future.wait(cache.map((fun) => fun()));
    return;
  }

  void close() {
    //_db.close();
  }
}

class Box<V> {
  final String name;
  final BoxCollection boxCollection;

  BoxBase? _cachedBox;

  Future<BoxBase> _getBox() async {
    return _cachedBox ??= await Hive.openLazyBox<V>(
      name,
      encryptionCipher: boxCollection._cipher,
    );
  }

  Box(this.name, this.boxCollection) {
    if (!(V is String ||
        V is bool ||
        V is int ||
        V is Object ||
        V is List<Object?> ||
        V is Map<String, Object?> ||
        V is double)) {
      throw ('Value type ${V.runtimeType} is not one of the allowed value types {String, int, double, List<Object?>, Map<String, Object?>}.');
    }
  }

  Future<List<String>> getAllKeys() async {
    final box = await _getBox();
    return box.keys.cast<String>().toList();
  }

  Future<Map<String, V>> getAllValues() async {
    final box = await _getBox();
    final keys = box.keys.toList();
    if (box is LazyBox) {
      final values = await Future.wait(keys.map(box.get));
      return {for (var i = 0; i < values.length; i++) keys[i]: values[i]};
    }
    return (box as h.Box).toMap().cast<String, V>();
  }

  Future<V?> get(String key) async {
    final box = await _getBox();
    if (box is LazyBox) return await box.get(key);
    return (box as h.Box).get(key);
  }

  Future<List<V?>> getAll(
    List<String> keys,
  ) async {
    final box = await _getBox();
    final values = <V?>[];
    for (final key in keys) {
      if (box is LazyBox) {
        values.add(await box.get(key));
      } else {
        values.add((box as h.Box).get(key));
      }
    }
    return values;
  }

  Future<void> put(String key, V val) async {
    final box = await _getBox();
    final future = () => box.put(key, val);
    if (boxCollection._txnCache != null) {
      boxCollection._txnCache?.add(future);
    } else {
      await future();
    }
    return;
  }

  Future<void> delete(String key) async {
    final box = await _getBox();
    final future = () => box.delete(key);
    if (boxCollection._txnCache != null) {
      boxCollection._txnCache?.add(future);
    } else {
      await future();
    }
    return;
  }

  Future<void> deleteAll(List<String> keys) async {
    final box = await _getBox();
    final future = () => box.deleteAll(keys);
    if (boxCollection._txnCache != null) {
      boxCollection._txnCache?.add(future);
    } else {
      await future();
    }
    return;
  }

  Future<void> clear() async {
    final box = await _getBox();
    final future = () => box.deleteAll(box.keys);
    if (boxCollection._txnCache != null) {
      boxCollection._txnCache?.add(future);
    } else {
      await future();
    }
    return;
  }
}
